/**
 * SocketThread (C)opyright 2007 Sebastian Kasparek
 *
 * more information is in the doc/SocketReadMe.txt file
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details. 
 *
 * You should have received a copy of the GNU Library General Public 
 * License along with this library; if not, write to the Free 
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA
 */

#include "socket.h"

SocketThread::SocketThread() : Looper( "goirc_socket_write" )
{
	// bind to the application object, which is the controller
	m_cMessenger = os::Messenger( os::Application::GetInstance() );
	m_hSocket = -1;
}

// FIXME: maybe use a macro here?
void SocketThread::PostErrorCode( const int8 nError )
{
	os::Message cMessage = os::Message( M_SOCKET_ERROR );
	cMessage.AddInt8( "errornumber", nError );
	m_cMessenger.SendMessage( &cMessage );
}

// FIXME: maybe use a macro here?
void SocketThread::PostStringMessage( const int nMessageCode, const char* pzName, const os::String* pcString )
{
	os::Message cMessage = os::Message( nMessageCode );
	cMessage.AddString( pzName, pcString->c_str() );
	m_cMessenger.SendMessage( &cMessage );
}

bool SocketThread::OpenSocket( const os::String* pcServer, const int nPort )
{
	// create the address and host structures
	struct sockaddr_in sAddress;
	sAddress.sin_family = AF_INET;
	struct hostent* psHost;

	// add the port (converted to the right endian)
	sAddress.sin_port = htons( nPort );

	// check if not a valid ip address (as it most often will be, when it is a servername),
	// if not valid: try to resolve as hostname
	// if valid: use it
	if( !inet_aton( pcServer->c_str(), &sAddress.sin_addr ) ) {
		psHost = gethostbyname( pcServer->c_str() );
		if( !psHost )
		{
			PostErrorCode( M_SOCKETERROR_SERVERNOTFOUND );
			m_hSocket = -1;
			return false;
		}
		
		// add the server ip
		sAddress.sin_addr = *(struct in_addr*)psHost->h_addr;
	}

	// server found, now save it
	m_cServer = pcServer->c_str();

	// open the socket
	m_hSocket = socket( PF_INET, SOCK_STREAM, 0 );
	if( m_hSocket < 0 ) {
		PostErrorCode( M_SOCKETERROR_COULDNOTOPEN );
		return false;
	}

	if( 0 != connect( m_hSocket, (struct sockaddr*)&sAddress, sizeof( struct sockaddr_in ) ) )
	{
		close( m_hSocket );
		PostErrorCode( M_SOCKETERROR_COULDNOTCONNECT );
		m_hSocket = -1;
		return false;
	}

	// connected to the server, so the port must be valid also, save it
	m_nPort = nPort;

	// start the socket read Thread
	m_pcSocketReadThread = new SocketReadThread( this, m_hSocket );
	m_pcSocketReadThread->Run();
	
	// initialize the Socket Read Messenger
	m_cMessengerSocketRead = os::Messenger( m_pcSocketReadThread );
	
	// tell the controller about the successful connection
	m_cMessenger.SendMessage( M_SOCKET_CONNECTED );

	// don't forget the SocketReader, only to come from the Handler into the Idle() Method
	// TODO: Maybe this is crap, and another way is prefered?
	m_cMessengerSocketRead.SendMessage( M_SOCKET_CONNECTED );

	return true;
}

void SocketThread::CloseSocket()
{
	if( m_hSocket >= 0 )
		close( m_hSocket );
	m_hSocket = -1;
}

void SocketThread::HandleMessage( os::Message* pcMessage )
{
	switch( pcMessage->GetCode() )
	{
		case M_SOCKET_CONNECT:
		{
			os::String cServer;
			int16 nPort;

			// get the connection destination
			pcMessage->FindString( "server", &cServer );
			pcMessage->FindInt16( "port", &nPort );
			
			PostStringMessage( M_SOCKET_INFO, "info", &os::String().Format( "connecting to server %s:%d", cServer.c_str(), nPort ) );

			if( !OpenSocket( &cServer, nPort ) )
				PostErrorCode( M_SOCKETERROR_COULDNOTCONNECT );

			PostStringMessage( M_SOCKET_INFO, "info", &os::String().Format( "connected to server %s:%d", cServer.c_str(), nPort ) );
		}
		break;
		
		case M_SOCKET_SEND:
		{
			int nBytes;

			// if socket is broken, send error message and return
			if( m_hSocket < 0 )
			{
				PostErrorCode( M_SOCKETERROR_DEAD );
				return;
			}
			
			// get the message to send
			os::String cString;
			while( 0 == pcMessage->FindString( "send", &cString ) )
			{
				pcMessage->RemoveData( "send" );
				// ignore empty messages
				if( cString.Length() < 1 )
					continue;

				// as we use a line based protocol (with irc, what this Class was written for),
				// add \r\n here and actually write to the socket
				nBytes = write( m_hSocket, os::String( cString + "\r\n" ).c_str(), strlen( cString.c_str() ) + 2 );
				if( nBytes == -1 )
				{
					PostErrorCode( M_SOCKETERROR_COULDNOTWRITE );
					// XXX: send the string back to the sender?
					continue;
				}
				else if( nBytes < static_cast<int>( strlen( cString.c_str() ) ) )
				{
					PostErrorCode( M_SOCKETERROR_NOTALLWRITTEN );
					// XXX: send the missing part of the string back to the sender?
					continue;
				}
	
				// tell the controller that the message was written to the socket, remove the \r\n
				PostStringMessage( M_SOCKET_SENT, "sent", &cString );
			}
		}
		break;
		
		case M_SOCKET_RECEIVED:
		{
			// don't do anything, only tell the controller what was received
			m_cMessenger.SendMessage( pcMessage );
		}
		break;
		
		case M_SOCKET_ERROR:
		case M_READSOCKET_ERROR:
		{
			m_cMessengerSocketRead.SendMessage( os::M_QUIT );
			CloseSocket();
		}
		break;
		
		default:
			std::cout << "not handled Message in SocketThread, really! " << pcMessage->GetCode() << std::endl;
			Handler::HandleMessage( pcMessage );
			break;
	}
}

bool SocketThread::OkToQuit()
{
	m_cMessengerSocketRead.SendMessage( os::M_QUIT );
	CloseSocket();
	return true;
}

SocketReadThread::SocketReadThread( const os::Messenger& cMessengerSocket, int hSocket ) : Looper( "goirc_read_socket" )
{
	m_cMessengerSocket = cMessengerSocket;
	m_hSocket = hSocket;
}

void SocketReadThread::PostError( int8 nError )
{
	os::Message cMessage( M_READSOCKET_ERROR );
	cMessage.AddInt8( "errornumber", nError );
	m_cMessengerSocket.SendMessage( &cMessage );
}	

bool SocketReadThread::Idle()
{
	// add the socket to the Filedescriptor Structure
	FD_ZERO( &m_sFDSet );
	FD_SET( m_hSocket, &m_sFDSet );
	// set maximum wait time to 1 second (give the Handler a chance to handle it's messages!)
	// TODO: probably this is crap, investigate a better way
	m_sTimeval.tv_sec = 1;
	m_sTimeval.tv_usec = 0;
	
	// wait for incoming data
	int i = select( m_hSocket + 1, &m_sFDSet, 0, 0, &m_sTimeval);

	// if select failed, but was not interrupted by a system call
	if( i < 0 && errno != EINTR )
	{
		PostError( M_SOCKETERROR_SELECTFAILED );
		// don't let the Handler call Idle() again, when he's done with his message
		return( false );
	}
	
	// if select has stopped blocking, check if there are data or if it timed out
	if( FD_ISSET( m_hSocket, &m_sFDSet ) )
	{
		// as we use a line based protocol (with irc, what this class was written for),
		// read one line from the socket
		unsigned int i = 0;
		char c;
		unsigned int nLength = sizeof( m_pzBuffer );
	
		while( c != '\n' && i < nLength )
		{
			if( read( m_hSocket, &c, sizeof( char ) ) != sizeof( char ) )
			{
				m_hSocket = -1;
				PostError( M_SOCKETERROR_DEAD );
				// don't let the Handler call Idle() again, when he's done with this message
				return( false );
			}
			m_pzBuffer[i++] = c;
		}
	
		m_pzBuffer[i<1?0:i-1] = 0;
		// remove '\r', if it was sent by the server as we don't need it
		if( i > 1 && m_pzBuffer[i-2] == '\r' ) m_pzBuffer[i-2] = 0;

		// tell the socket thread about received data
		os::Message cMessage( M_SOCKET_RECEIVED );
		cMessage.AddString( "received", m_pzBuffer );
		m_cMessengerSocket.SendMessage( &cMessage );
	}

	// let the Handler handle it's messages and then call Idle() again
	return true;
}

void SocketReadThread::HandleMessage( os::Message* pcMessage )
{
	if( pcMessage->GetCode() == M_SOCKET_CONNECTED ) return;
	
	std::cout << "not handled message in SocketReadThread: " << pcMessage->GetCode() << std::endl;
	Handler::HandleMessage( pcMessage );
}

